import numpy as np
import pandas as pd

from datawhys.datautils import apply_rule


def bias(df: pd.DataFrame, bias_variables: list = None, rule: dict = None,) -> dict:
    """Score a rule to detect bias dataframe. Used to determine potential issues with
    fairness.

    Parameters
    ----------
    df : pd.DataFrame
        The dataset to score. If rule is set, then df should represent the original
        population. If population is set, then df should represent the sample to score.

    bias_variables : list, default=None
        Which columns to use as the columns to evaluate for potential bias

    rule: dict, default=None
        The solver rule to use for scoring. This or population must be set, but not
        both. If set, df should be the population dataset.

    Returns
    -------
    bias_scores: dict
        A dictionary of the shift in each of the selected bias variables

        For continuous variables, shows the mean, std, min, max, and percentiles
        For categorical variables, shows the deltas and ratios of classes between
        the population and sample segments

    Examples
    --------
    >>> from datawhys import metrics
    >>> from datawhys.datasets import load_titanic
    >>> df = load_titanic()
    >>> bias_variables = ['sex', 'age']
    >>> rule = {
        'class': {'lo': 2, 'hi': 3}, 'sex': 'male',
        'parch': {'lo': 0, 'hi': 0},
        'ticketnumber': {'lo': 2151.0, 'hi': 3101280.0}
        }
    >>> datawhys.metrics.bias(df, bias_variables, rule)
    {
        "sex": {
            "counts_population": {
                "male": 842,
                "female": 466
            },
            "ratio_population": {
                "male": 0.6437308868501529,
                "female": 0.3562691131498471
            },
            "counts_sample": {
                "male": 513,
                "female": 0
            },
            "ratio_sample": {
                "male": 1,
                "female": 0
            },
            "ratio_difference": {
                "male": 0.3562691131498471,
                "female": -0.3562691131498471
            },
            "ratio_percent_change": {
                "male": 55.344418052256536,
                "female": -100
            }
        },
        "age": {
            "description_population": {
                "count": 1045,
                "mean": 29.881977703349285,
                "std": 14.420375254531496,
                "min": 0.1667,
                "25%": 21,
                "50%": 28,
                "75%": 39,
                "max": 80
            },
            "description_sample": {
                "count": 383,
                "mean": 29.66318537859008,
                "std": 10.92394589950367,
                "min": 11,
                "25%": 22,
                "50%": 27,
                "75%": 34,
                "max": 74
            },
            "description_delta": {
                "count": -662,
                "mean": -0.21879232475920674,
                "std": -3.4964293550278267,
                "min": 10.8333,
                "25%": 1,
                "50%": -1,
                "75%": -5,
                "max": -6
            },
            "description_ratio": {
                "count": -0.6334928229665071,
                "mean": -0.007321882337616659,
                "std": -0.2424645193563253,
                "min": 64.98680263947212,
                "25%": 0.04761904761904767,
                "50%": -0.0357142857142857,
                "75%": -0.1282051282051282,
                "max": -0.07499999999999996
            }
        }
    }

    """
    if rule is not None:
        population = df
        sample = apply_rule(df, rule)
    else:
        return TypeError("Must include a rule")

    ret_dict = {}

    for v in bias_variables:

        # check if bias variable is numeric or not
        if np.issubdtype(population[v].dtype, np.number):
            pop_cont = population[v].describe()
            sample_cont = sample[v].describe()
            # create return dictionary
            ret_dict[v] = {
                "description_population": pop_cont.to_dict(),
                "description_sample": sample_cont.to_dict(),
                "description_delta": (sample_cont - pop_cont).to_dict(),
                # subtract one to have +/- percentages
                "description_ratio": (sample_cont / pop_cont - 1).to_dict(),
            }
        else:
            # create return dataframe and then set column name
            ret_df = pd.DataFrame(population[v].value_counts())
            ret_df.columns = ["counts_population"]

            ret_df["ratio_population"] = (
                ret_df["counts_population"] / ret_df["counts_population"].sum()
            )
            ret_df["counts_sample"] = pd.DataFrame(sample[v].value_counts())
            ret_df["ratio_sample"] = (
                ret_df["counts_sample"] / ret_df["counts_sample"].sum()
            )
            ret_df = ret_df.fillna(0)
            ret_df["ratio_difference"] = (
                ret_df["ratio_sample"] - ret_df["ratio_population"]
            )
            ret_df["ratio_percent_change"] = (
                ret_df["ratio_sample"] / ret_df["ratio_population"] - 1.0
            ) * 100.0

            ret_dict[v] = ret_df.to_dict()

    return ret_dict



from typing import Dict, Union

import numpy as np
import pandas as pd
from scipy.stats import norm
from sklearn.preprocessing import StandardScaler


def data_drift(
    model_data: Union[pd.DataFrame, pd.Series, np.ndarray],
    test_data: Union[pd.DataFrame, pd.Series, np.ndarray],
    return_components: bool = False,
) -> Dict:
    """Test if there is data drift from the data a model was train on and any
        new unseen data. The test works only for continuous, non-binary data.

        Parameters
        ----------
        model_data : pd.DataFrame, pd.Series, np.ndarray
            The data a model was used to train on. The model will have weights or
            coefficients that are tuned to this data and its underlying distribution.

        test_data : pd.DataFrame, pd.Series, np.ndarray
            The new data your model has not yet seen. This data may have a distribution
            that is significantly different from the model data and may explain
            unexpected model results.

        return_components : boolean, default=False
            For multivariate data sets, return the component for both the model data and
            test data discovered by UMAP. For univariate data sets, the same data will
            be returned.

        Returns
        -------
        test_metrics: dict
            z-score of the test,
            p-value of the test,
            confidence of the test

            if returning the components, both model and test data components
            will be returned

        Examples
        --------
        >>> mondobrain.metrics.data_drift(model_data, test_data)
        {
            "z-score": -0.7067833851248948,
            "p-value": 0.7601494493749683,
            "confidence": 0.2398505506250317
        }

        In this example, there was not signficant indication that the test data
        was different than the model data. Any unexpected results of the model
        are likely not due to data drift with a confidence of 76% in this case
        due to rejection of the null hypothesis with insignificant p-value.
    """

    import umap

    if model_data.shape[1] != test_data.shape[1]:
        raise ValueError("Model and Test data must have same number of dimensions")

    if model_data.shape[0] > 1:
        df = np.concatenate((model_data, test_data), axis=0)
        scaler = StandardScaler()
        df = scaler.fit_transform(df)
        df_model = df[: model_data.shape[0]]
        df_test = df[model_data.shape[0] :]

        embedding_train = umap.UMAP(
            n_neighbors=5, min_dist=0.0, n_components=1, random_state=42,
        ).fit_transform(df_model)

        embedding_test = umap.UMAP(
            n_neighbors=5, min_dist=0.0, n_components=1, random_state=42,
        ).fit_transform(df_test)
    else:
        embedding_train = model_data
        embedding_test = test_data

    model_mean = embedding_train.mean()
    test_mean = embedding_test.mean()
    model_std = embedding_train.std()
    test_sqrt_n = np.sqrt(embedding_test.shape[0])

    z_score = test_sqrt_n * (test_mean - model_mean) / model_std
    p = norm.sf(z_score)
    confidence = 1 - p

    return_package = {"z_score": z_score, "p-value": p, "confidence": confidence}

    if return_components:
        return_package["components"] = {
            "model": embedding_train.tolist(),
            "test": embedding_test.tolist(),
        }

    return return_package



from ._apply import apply_rule, apply_rule_tree, apply_rules
from ._decorrelate import decorrelate
from ._sample import sample

__all__ = ["apply_rule", "apply_rules", "apply_rule_tree", "sample", "decorrelate"]



# flake8: noqa
import logging

__author__ = "DataWhys"
__version__ = "0.3.1"

# Configuration variables
auth_enabled = True
api_key = None
api_secret = None
api_token = None

api_base = "https://api.datawhys.ai/"
auth0_domain = "mb-production.us.auth0.com"

verify_ssl_certs = True
proxy = None
default_http_client = None
enable_telemetry = True
max_network_retries = 0
ca_bundle_path = None

# Set to either 'debug' or 'info', controls console logging
log = None

from datawhys import api  # isort:skip
from datawhys.core.api import DataWhysFrame, DataWhysSeries  # isort:skip
from datawhys.prescriber import Solver  # isort:skip



from typing import Dict, List, Union

import numpy as np
import pandas as pd

from datawhys.utils.graph import create_rule_graph


def _condition_mask(s: pd.Series, cond: Union[dict, str]):
    if isinstance(cond, dict):
        lo, hi = cond["lo"], cond["hi"]
        return (s >= lo) & (s <= hi)

    return s == cond


def conditions_mask(df: Union[pd.Series, pd.DataFrame], rule: dict):
    """ Get a mask based on a dictionary of conditions
    """
    masks = []
    for key, cond in rule.items():
        masks.append(_condition_mask(df[key], cond))

    if isinstance(df, pd.Series):
        return (
            np.ones(df.size, dtype=np.bool)
            if np.all(masks)
            else np.zeros(df.size, dtype=np.bool)
        )
    elif df.shape[0] == 1:
        return np.array([np.all([mask.values for mask in masks])])
    else:
        return np.all(np.array(masks), axis=0)


def apply_rule(
    df: Union[pd.Series, pd.DataFrame], rule: dict, inverse: bool = False
) -> pd.DataFrame:
    """ Applies the rule to the population df to obtain a sample

    Parameters
    ----------
    df : pd.DataFrame
        The dataset to apply the rule to

    rule : dict
        A set of conditions to apply to the dataframe

    inverse : boolean, default=False
        If True, the inverse of the rule is applied

    Returns
    -------
    sample: pd.DataFrame
        The sample of the original dataframe based on rule conditions
    """
    if not rule:
        raise ValueError("An empty rule cannot be applied to a dataframe")

    mask = conditions_mask(df, rule)

    if inverse:
        mask = ~mask

    return df[mask]


def apply_rules(
    df: pd.DataFrame, rules: List[Dict], inverse: bool = False, operation: str = "or"
) -> pd.DataFrame:
    """ Applies a set of rules to the population df to obtain a sample

    Parameters
    ----------
    df : pd.DataFrame
        The dataset to apply the rule to

    rule : List collection of Dict
        A list of a set of conditions to apply to the dataframe

    inverse : boolean, default=False
        If True, the inverse of the rule is applied

    operation : str, default="or"
        Options are "or" or "and"
        If "or", the union of the rules is applied
        If "and", the intersection of the rules is applied

    Returns
    -------
    sample: pd.DataFrame
        The sample of the original dataframe based on rule conditions
    """
    mask = [conditions_mask(df, rule) for rule in rules if rule]

    if len(mask) == 0:
        raise ValueError("No rule found to apply to the dataframe")

    if operation == "or":
        mask = np.any(mask, axis=0)
    elif operation == "and":
        mask = np.all(mask, axis=0)
    else:
        raise ValueError("operation type not supported")

    if inverse:
        mask = ~mask

    return df[mask]


def apply_rule_tree(
    df: pd.DataFrame, rule_tree: Dict, inverse: bool = False
) -> pd.DataFrame:
    """ Applies a set of hierarchical rules to the population df to obtain a sample

    Parameters
    ----------
    df : pd.DataFrame
        The dataset to apply the rule to

    rule_tree : List collection of Dict
        A hierarchical set of conditions to apply to the dataframe

    inverse : boolean, default=False
        If True, the inverse of the rule is applied

    Returns
    -------
    sample: pd.DataFrame
        The sample of the original dataframe based on rule conditions
    """
    G = create_rule_graph(rule_tree)

    df_dict = dict()

    def segment_dataset(starting_node, df):
        rule = G.nodes[starting_node]["rule"]
        try:
            df_in = apply_rule(df, rule)
        except KeyError:
            df_in = df
        try:
            df_out = apply_rule(df, rule, inverse=True)
        except KeyError:
            df_out = df.loc[[]]
        neighbors = [
            node for node in G.neighbors(starting_node) if node != starting_node
        ]
        if neighbors:
            for node in neighbors:
                if node[-1] == "IN":
                    segment_dataset(node, df_in)
                elif node[-1] == "OUT":
                    segment_dataset(node, df_out)
        else:
            previous_level = starting_node[-2]
            current_level = previous_level + 1
            df_dict[starting_node + (current_level, "IN")] = df_in
            df_dict[starting_node + (current_level, "OUT")] = df_out

    segment_dataset((0, "IN", 0, "IN"), df)

    point_index = pd.concat(
        [df_dict[node] for node in df_dict if node[-1] == "IN"]
    ).index

    mask = np.in1d(df.index, point_index)
    if inverse:
        mask = ~mask

    return df[mask]



import numpy as np
import pandas as pd


def target_metrics(s: pd.Series, target: str = None) -> dict:
    """Get metrics for a target within a column

    Parameters
    ----------
    s : pd.Series
        The series to calculate metrics for

    target: str, default=None
        The class in s to calculate metrics for. For a continuous outcome this should
        either be `min` or `max`. If None, then the first class or `max` is selected.

    Examples
    --------
    >>> from datawhys.datasets import load_titanic
    >>> from datawhys import metrics
    >>> df = load_titanic()
    >>> series = df['survived']
    >>> target = "Died"
    >>> metrics.target_metrics(series, target)
    {
        'mean': 0.617737003058104,
        'std': 0.48594032361072487,
        'size': 1308,
        'size_above_mean': 808
    }
    """
    if target is None:
        target = s[0] if s.dtype == np.object else "max"

    if s.dtype == np.object:
        values = (s == target).astype(np.int)
    else:
        values = s
        if target == "min":
            values *= -1

    mean = values.mean() if values.size > 0 else 0
    above_mean = int((values >= mean).sum()) if mean > 0 else 0

    return {
        "mean": mean,
        "std": values.std(ddof=0) if values.size > 0 else 0,
        "size": int(values.size),
        "size_above_mean": above_mean,
    }



"""
The :mod:`datawhys.solver.meta` submodule contains solve methods for
meta solves.
"""

from ._solve import class_exploration_solve, exclusion_set_solve

__all__ = ["class_exploration_solve", "exclusion_set_solve"]



import numpy as np
import pandas as pd

from datawhys.datautils import apply_rule
from datawhys.dd_transformer import DDTransformer
from datawhys.utils import utilities
from datawhys.utils.data import is_continuous

from ._target import target_metrics


def score(
    df: pd.DataFrame,
    outcome: str = None,
    target: str = None,
    rule: dict = None,
    population: pd.DataFrame = None,
):
    """Score a dataframe

    Parameters
    ----------
    df : pd.DataFrame
        The dataset to score. If rule is set, then df should represent the original
        population. If population is set, then df should represent the sample to score.

    outcome : str, default=None
        Which column to use as the outcome for the score (sometimes referred to as a
        target feature). If None, then the first column is selected

    target: str, default=None
        The class in the outcome column to target. For a continuous outcome this should
        either be `min` or `max`. If None, then the first class or `max` is selected.

    rule: dict, default=None
        The solver rule to use for scoring. This or population must be set, but not
        both. If set, df should be the population dataset

    population: pd.DataFrame, default=None
        The population dataset to base the score off of. If set, df should be the sample
        dataset.

    Returns
    -------
    score: float
        The score of the sample and population

    Examples
    --------
    >>> from datawhys.datasets import load_titanic
    >>> from datawhys import metrics
    >>> df = load_titanic()
    >>> outcome = 'survived'
    >>> target = "Died"
    >>> rule = {
        'class': {'lo': 2, 'hi': 3},
        'sex': 'male',
        'parch': {'lo': 0, 'hi': 0},
        'ticketnumber': {'lo': 2151.0, 'hi': 3101280.0}
        }
    >>> metrics.score(df, outcome, target, rule)
    13.001719638805223
    """
    # Quick checks to make sure that one of rule and population is set.
    if rule is None and population is None:
        raise ValueError("one of `rule` or `population` must be set")

    if rule is not None and population is not None:
        raise ValueError("only one of `rule` or `population` may be set")

    if outcome is None:
        outcome = df.columns[0]

    if target is None:
        target = df[outcome].iloc[0] if df[outcome].dtype == np.object else "max"

    if population is not None:
        sample = df

    if rule is not None:
        population = df
        sample = apply_rule(df, rule)

    # Encode our outcome column if the outcome column is continuous
    if is_continuous(population[outcome].dtype):
        encoder = DDTransformer(population)
        population = utilities.encode_dataframe(population, encoder)
        sample = utilities.encode_dataframe(sample, encoder)

        # Encode outcome. Target not needed as this is continuous
        outcome = encoder.original_to_encoded_column(outcome)

    # Get metrics for population and sample
    pop_metrics = target_metrics(population[outcome], target)
    smp_metrics = target_metrics(sample[outcome], target)

    # Return 0 if we our size is 0
    if (
        smp_metrics["size"] == 0
        or pop_metrics["size"] == 0
        or smp_metrics["mean"] == pop_metrics["mean"]
        or pop_metrics["std"] == 0
    ):
        return 0

    return (
        np.sqrt(smp_metrics["size"])
        * (smp_metrics["mean"] - pop_metrics["mean"])
        / pop_metrics["std"]
    )


def _exclude_key(a: dict, key) -> dict:
    return {k: v for k, v in a.items() if k != key}


def partial_scores(
    df: pd.DataFrame, rule: dict, outcome: str = None, target: str = None,
) -> dict:
    """Calculates the partial score for each condition in a rule.

    Parameters
    ----------
    df : pd.DataFrame
        The dataset to score. The df should represent the original
        population.

    rule: dict
        The solver rule to use for scoring.

    outcome : str, default=None
        Which column to use as the outcome for the score (sometimes referred to as a
        target feature). If None, then the first column is selected

    target: str, default=None
        The class in the outcome column to target. For a continuous outcome this should
        either be `min` or `max`. If None, then the first class or `max` is selected.


    Returns
    -------
    scores: dict
        A dictionary of scores for each key/condition in the rule


    Examples
    --------
    >>> from datawhys.datasets import load_titanic
    >>> from datawhys import metrics
    >>> df = load_titanic()
    >>> outcome = 'sex'
    >>> target = 'male'
    >>> rule = {
            'class': {'lo': 2, 'hi': 3},
            'sex': 'male',
            'parch': {'lo': 0, 'hi': 0},
            'ticketnumber': {'lo': 2151.0, 'hi': 3101280.0}
        }
    >>> metrics.partial_scores(df, rule, outcome, target)
    {
        'class': -2.175216344281214,
        'sex': 11.817078598704317,
        'parch': -1.2967726194944724,
        'ticketnumber': -0.8020530610406418
    }
    """
    if not rule:
        raise ValueError("An empty rule is not permitted")

    base_score = score(df, outcome=outcome, target=target, rule=rule)

    if len(rule.keys()) == 1:
        key = list(rule.keys())[0]
        return {key: base_score}

    scores = {}
    for key, cond in rule.items():
        subrule = _exclude_key(rule, key)
        subscore = score(df, outcome=outcome, target=target, rule=subrule)

        scores[key] = base_score - subscore

    return scores


def feature_signals(
    df: pd.DataFrame, outcome: str = None, target: str = None, **solve_params,
):
    """
    Calculate the signal/strength of each variable in a dataframe.

    Parameters
    ----------
    df: Pandas DataFrame
        Dataframe containing the Outcome Feature and Independent Feature variables.

    outcome: str (optional), default=None
        The label of the dependent variable used for learning.  If None, the first
        column is used.

    target: str (optional), default=None
        The target class of the outcome if discrete.  'min' or 'max' if continuous.

    **solve_params: any
        Parameters to pass on to `solve`

    Returns
    -------
    signals: dict
        A dictionary mapping each variable with their score and condition
        leading to the score.  This is achieved by doing a one dimensional
        solve on each variable.

    Examples
    --------
    >>> from datawhys.datasets import load_titanic
    >>> df = load_titanic()
    >>> metrics.feature_signals(df, 'survived', "Died")
    {
        'sex': {'condition': 'male', 'score': 11.408374681094228},
        'fare': {'condition': {'lo': 0.0, 'hi': 10.4625}, 'score': 7.271397191659396},
        'class': {'condition': {'lo': 3, 'hi': 3}, 'score': 6.932868263529628},
        'ticketnumber': {'condition': {'lo': 345763.0, 'hi': 366713.0},
        'score': 6.411057911385609},
        'embarked': {'condition': 'S', 'score': 3.0651604266866816},
        'parch': {'condition': {'lo': 0, 'hi': 0}, 'score': 3.033943859578286},
        'sibsp': {'condition': {'lo': 0, 'hi': 0}, 'score': 2.153140749023159},
        'age': {'condition': {'lo': 17.0, 'hi': 28.5}, 'score': 0.9015429873144141},
        'name': {'condition': {}, 'score': 0},
        'ticket': {'condition': {}, 'score': 0},
        'ticketcode': {'condition': {}, 'score': 0},
        'cabin': {'condition': {}, 'score': 0},
        'cabinsection': {'condition': {}, 'score': 0},
        'boat': {'condition': {}, 'score': 0},
        'body': {'condition': {}, 'score': 0},
        'homedest': {'condition': {}, 'score': 0}
    }
    """
    from ..solver import solve

    if not outcome:
        outcome = df.columns[0]

    results = {}

    for col in df.columns:
        if col == outcome:
            continue

        df_one_dim = df[[col, outcome]]

        rule = solve(df_one_dim, outcome=outcome, target=target, **solve_params)
        rule_score = score(df, outcome, target, rule) if rule else 0

        results[col] = {"condition": rule[col] if rule else {}, "score": rule_score}

    sorted_results = {
        k: v
        for k, v in sorted(results.items(), key=lambda x: x[1]["score"], reverse=True)
    }

    return sorted_results



import json
from typing import List

import pandas as pd

from datawhys import datautils
from datawhys.api import meta_class_start, meta_exclusion_set_start, meta_result
from datawhys.dd_transformer import DDTransformer
from datawhys.utils import meta as meta_utils, utilities


def _prepare_data(
    df: pd.DataFrame,
    outcome: str,
    classes: List[str],
    encode: bool,
    sample: bool,
    random_state,
):
    if outcome is None:
        outcome = df.columns[0]

    if encode:
        df_orig = df.copy()
        outcome_orig = outcome

        encoder = DDTransformer(df)
        df = utilities.encode_dataframe(df, encoder)
        outcome = encoder.original_to_encoded_column(outcome)

        if classes:
            classes = [
                utilities.encode_value(df_orig, outcome_orig, c, encoder)
                for c in classes
            ]
    else:
        encoder = None

    if sample:
        df = datautils.sample(
            df, 2500, outcome=outcome, floor=3500, random_state=random_state,
        )

    return df, outcome, classes, encoder


def _postprocess_data(
    df: pd.DataFrame, outcome: str, encoder: DDTransformer, rules: dict
):
    if encoder:
        rules = meta_utils.decode_meta_rules(rules, df, outcome, encoder)

    return rules


def class_exploration_solve(
    df: pd.DataFrame,
    outcome: str = None,
    classes: List[str] = None,
    encode=True,
    sample=True,
    random_state=None,
    **kwargs,
):
    """Run a solve that learns against itself

    Parameters
    ----------
    df : pd.DataFrame
        The dataset to use for the solve

    outcome : str, default=None
        Which column to use as the outcome for the solve (sometimes referred to as a
        target feature). If None, then the first column is selected

    classes: List[str], default=None
        For discrete outcomes, a list of modalities to explore. For continuous outcomes,
        a list containing 'min' and/or 'max'. If None, explore all classes.

    encode: bool, default=True
        Whether or not the data should be encoded before being sent to the MB API.
        Encoding can result in additional time on client side. Disable if your data is
        largely non-sensitive.

    sample: bool, default=True
        Whether or not the data should be sampled before being sent to the MB API.
        Not pre-sampling the data can cause size limits to be reached and excessive
        solve times

    random_state : int or np.random.RandomStateInstance, default: 0
        Pseudo-random number generator to control the sampling state.
        Use an int for reproducible results across function calls.
        See the sklearn for more details.

    kwargs: any
        Remaining kwargs are sent so api.meta_class_start

    Returns
    -------
    rules: array of dict
        An array of meta_rule dicts that have rules and meta properties
    """
    df, outcome, classes, encoder = _prepare_data(
        df, outcome, classes, encode, sample, random_state
    )

    # Send data OTW and wait for solve to finish
    data = json.loads(df.to_json(orient="records"))
    task = meta_class_start(outcome=outcome, data=data, classes=classes, **kwargs)
    result = meta_result(id=task["id"])

    rules = result["rules"]
    rules = _postprocess_data(df, outcome, encoder, rules)

    return rules


def exclusion_set_solve(
    df: pd.DataFrame,
    outcome: str = None,
    classes: List[str] = None,
    encode=True,
    sample=True,
    random_state=None,
    **kwargs,
):
    """Run a solve that learns against itself

    Parameters
    ----------
    df : pd.DataFrame
        The dataset to use for the solve

    outcome : str, default=None
        Which column to use as the outcome for the solve (sometimes referred to as a
        target feature). If None, then the first column is selected

    classes: List[str], default=None
        For discrete outcomes, a list of modalities to explore. For continuous outcomes,
        a list containing 'min' and/or 'max'. If None, explore all classes.

    encode: bool, default=True
        Whether or not the data should be encoded before being sent to the MB API.
        Encoding can result in additional time on client side. Disable if your data is
        largely non-sensitive.

    sample: bool, default=True
        Whether or not the data should be sampled before being sent to the MB API.
        Not pre-sampling the data can cause size limits to be reached and excessive
        solve times

    random_state : int or np.random.RandomStateInstance, default: 0
        Pseudo-random number generator to control the sampling state.
        Use an int for reproducible results across function calls.
        See the sklearn for more details.

    kwargs: any
        Remaining kwargs are sent so api.meta_exclusion_set_start

    Returns
    -------
    rules: array of dict
        An array of meta_rule dicts that have rules and meta properties
    """
    df, outcome, classes, encoder = _prepare_data(
        df, outcome, classes, encode, sample, random_state
    )

    # Send data OTW and wait for solve to finish
    data = json.loads(df.to_json(orient="records"))
    task = meta_exclusion_set_start(
        outcome=outcome, data=data, classes=classes, **kwargs
    )
    result = meta_result(id=task["id"])

    rules = result["rules"]
    rules = _postprocess_data(df, outcome, encoder, rules)

    return rules



"""
The :mod:`datawhys.solver` module includes pure solves, ensemble-like solves
and meta solve capabilities
"""

from ._solver import exhaustive_solve, solve

__all__ = ["exhaustive_solve", "solve"]



from typing import List, Mapping

import numpy as np
import pandas as pd
from sklearn.preprocessing import LabelEncoder

from datawhys.utils.data import is_discrete

Matrix = Mapping[str, Mapping[str, float]]


def correlation_matrix(
    df: pd.DataFrame, method: str = "spearman", encode: bool = True
) -> Matrix:
    """
    Compute the correlation matrix of each column in the dataframe.

    Parameters
    ----------
    df: Pandas DataFrame
        Dataframe containing variables that you want to calculate the correlation of.

    method: str {'spearman', 'tau', 'pearson'}, default='spearman'
        Correlation algorithm to use

    encode: bool, default=True
        If True, categorical variables are one hot encoded and the maximum correlation
        of a category is taken.  If False, no encoding takes place.

    Returns
    -------
    corr_matrix: dict
        A nested dictionary that maps variable -> other variable -> correlation value
        between -1 and 1, inclusive.

    Examples
    --------
    >>> from datawhys.datasets import load_iris
    >>> from datawhys import metrics
    >>> df = load_iris()
    >>> metrics.correlation_matrix(df)
    array([[2, 0, 0],
           [0, 0, 1],
           [1, 0, 2]])
    >>> y_true = ["cat", "ant", "cat", "cat", "ant", "bird"]
    >>> y_pred = ["ant", "ant", "cat", "cat", "ant", "cat"]
    >>> confusion_matrix(y_true, y_pred, labels=["ant", "bird", "cat"])
    {
        "SepalLength": {
            "PetalLength": 0.8813863932886515,
            "PetalWidth": 0.8344206519767946,
            "SepalWidth": -0.15945651848582867
        },
        "SepalWidth": {
            "PetalLength": -0.3034206463815158,
            "PetalWidth": -0.2775110724763029,
            "SepalLength": -0.15945651848582867
        },
        "PetalLength": {
            "PetalWidth": 0.9360033509355781,
            "SepalLength": 0.8813863932886515,
            "SepalWidth": -0.3034206463815158
        },
        "PetalWidth": {
            "PetalLength": 0.9360033509355781,
            "SepalLength": 0.8344206519767946,
            "SepalWidth": -0.2775110724763029
        }
    }

    """

    def one_hot_enc_label(value):
        for ohe_col in one_hot_enc_cols:
            if value.startswith(ohe_col):
                return ohe_col
        return value

    def abs_max(series):
        if series.size == 0:
            return 0

        col = series.sort_values()
        a, b = col[0], col[col.size - 1]

        return a if np.abs(a) >= np.abs(b) else b

    df_copy = df.copy()

    # convert to str to ensure that sortable types. Sometimes dates mess this up
    for col in df_copy.columns:
        if is_discrete(df[col]):
            df_copy[col] = df_copy[col].astype(str)

    # using DDTransformer as encoder drops columns, so use LabelEncoder instead
    edf = df_copy.apply(LabelEncoder().fit_transform)
    one_hot_enc_cols = [
        col
        for col in edf.columns
        if is_discrete(edf[col]) and len(edf[col].unique()) <= 12
    ]

    if encode:
        edf = pd.get_dummies(edf, columns=one_hot_enc_cols)

    df_corr = edf.corr(method=method)
    df_corr = df_corr.fillna(0)

    if encode:
        df_corr = df_corr.groupby(one_hot_enc_label, axis=0).agg(abs_max)
        df_corr = df_corr.groupby(one_hot_enc_label, axis=1).agg(
            lambda x: x.agg(abs_max, axis=1)
        )

    corr_matrix = df_corr.to_dict()
    corr_matrix = {
        out_var: {
            in_var: corr
            for in_var, corr in sorted(map.items(), key=lambda x: -np.abs(x[1]))
            if out_var != in_var
        }
        for out_var, map in corr_matrix.items()
    }

    return corr_matrix


def correlation_groups(
    corr_matrix: Matrix, outcome: str = None, threshold: float = 0.5
) -> List[dict]:
    """
    Build groups of correlated variables based on their correlation matrix.

    Parameters
    ----------
    corr_matrix: Matrix
        A correlation matrix built by the `correlation_matrix` function.

    outcome: str, default=None
        If given, variables are sorted by their correlation to the outcome before
        grouping.  This ensures that variables with the highest importance with respect
        to the outcome are group heads.

    threshold: float, default=0.5
        Minimum threshold to assign a variable to a group.  If the correlation between
        a variable and a group head is at least the threshold value, then the variable
        is placed in that group.

    Returns
    -------
    groups: List[dict]
        A list of groups, each with a head variable and a sub-list of variables in
        that group.

    Examples
    --------
    >>> from datawhys import metrics
    >>> from datawhys.datasets import load_iris
    >>> df = load_iris()
    >>> corr_matrix = metrics.correlation_matrix(df)
    >>> metrics.correlation_groups(corr_matrix)
    [
        {'name': 'Name',
        'variables': [{'name': 'PetalWidth',
        'correlationToHead': 0.9378500422381613},
        {'name': 'PetalLength', 'correlationToHead': 0.9354413500363797},
        {'name': 'SepalLength', 'correlationToHead': 0.7980781172420549}]},
        {'name': 'SepalWidth', 'variables': []}
    ]
    """
    if threshold < 0 or threshold > 1:
        raise ValueError("Threshold must be a float between 0 and 1, inclusive.")

    groups = []

    if outcome:
        vars = sorted(
            (k for k in corr_matrix.keys() if k != outcome),
            key=lambda x: -np.abs(corr_matrix[x][outcome]),
        )
    else:
        vars = corr_matrix.keys()
    used_vars = {x: False for x in vars}

    for head in vars:
        if used_vars[head]:
            continue

        sub_groups = []

        for sub, corr in corr_matrix[head].items():
            if (outcome and sub == outcome) or used_vars[sub]:
                continue
            if np.abs(corr) < threshold:
                break

            used_vars[sub] = True
            sub_data = {
                "name": sub,
                "correlationToHead": corr,
            }
            if outcome:
                sub_data.update({"correlationToOutcome": corr_matrix[sub][outcome]})
            sub_groups.append(sub_data)

        head_data = {"name": head, "variables": sub_groups}
        if outcome:
            head_data.update({"correlationToOutcome": corr_matrix[head][outcome]})

        groups.append(head_data)

    return groups



import networkx as nx


def create_rule_graph(rules):
    G = nx.DiGraph()

    for node in rules:
        lbl = "score: {:f}\nsize: {}".format(rules[node]["score"], rules[node]["size"])
        G.add_node(node, rule=rules[node]["rule"], label=lbl)

    for node in G.nodes:
        if node[:-2] in G.nodes:
            lbl = node[-1]
            G.add_edge(node[:-2], node, label=lbl)

    return G



import pandas as pd


def _decorrelate_matrix(
    df: pd.DataFrame,
    method: str = "spearman",
    encode: bool = True,
    outcome: str = None,
    threshold: float = 0.5,
) -> pd.DataFrame:

    from datawhys.metrics import correlation_groups, correlation_matrix

    corr_matrix = correlation_matrix(df, method=method, encode=encode)
    groups = correlation_groups(corr_matrix, outcome=outcome, threshold=threshold)

    cols = []
    if outcome:
        cols.append(outcome)
    cols += [x["name"] for x in groups]

    return df[cols] if len(cols) > 0 else df


def _decorrelate_signals(
    df: pd.DataFrame,
    threshold: [float, float] = None,
    outcome: str = None,
    target: str = None,
    **solve_params
) -> pd.DataFrame:

    from datawhys.metrics import feature_signals

    if threshold is not None and (
        len(threshold) != 2
        or (
            threshold[0] is not None
            and threshold[1] is not None
            and threshold[0] >= threshold[1]
        )
    ):
        raise ValueError(
            "Threshold must be [min, max] where min < max, [None, max] or [min, None]."
        )

    signals = feature_signals(df, outcome=outcome, target=target, **solve_params)

    cols = []
    if outcome:
        cols.append(outcome)

    for col, data in signals.items():
        if threshold is not None:
            if threshold[0] is not None and data["score"] < threshold[0]:
                continue
            if threshold[1] is not None and data["score"] > threshold[1]:
                continue

        cols.append(col)

    return df[cols] if len(cols) > 0 else df


def decorrelate(df: pd.DataFrame, technique: str = "signals", **kwargs) -> pd.DataFrame:
    """
    Remove highly correlated columns in a dataframe.

    Parameters
    ----------
    df: Pandas DataFrame
        Dataframe containing variables that you want to decorrelate.

    technique: str {'signals', 'matrix'}, default='signals'
        Decorrelation technique used.
        If 'signals', it uses the datawhys.metrics.feature_signals function to remove
        feature variables that have a score below a specified threshold.
        If 'matrix', it uses the datawhys.metrics.correlation_matrix function to
        selection correlation group heads based on a specified threshold.

    **kwargs
        Arguments to pass to the corresponding functions in the datawhys.metrics
        package.

    Returns
    -------
    decorrelated_df: Pandas Dataframe
        A dataframe with the columns correlated above the specified threshold
        removed.
    """
    if technique == "signals":
        return _decorrelate_signals(df, **kwargs)
    elif technique == "matrix":
        return _decorrelate_matrix(df, **kwargs)
    else:
        raise ValueError("Invalid technique '%s' given." % technique)



from typing import List

import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler

from datawhys.dd_transformer import DDTransformer
from datawhys.solver import solve
from datawhys.utils import utilities
from datawhys.utils.data import is_continuous

__X_COL = "__transform_x"
__Y_COL = "__transform_y"
__DIST_COL = "__point_distance"
__INV_DIST_COL = "__point_inverse_distance"


def find_adjustable_conditions(
    df: pd.DataFrame,
    point: dict,
    adjustable: List[str],
    inverse: bool = True,
    outcome: str = None,
    targets: List[str] = None,
    **kwargs,
):
    """
    Find a rule using What-If Analysis on a given point and a dataset.  Based on the
    point evaluation, finds conditions from the list of adjustable columns for similar
    or different points in the dataset.

    Parameters
    ----------
    df: pd.DataFrame
        The dataset to use for the analysis

    point: dict
        A single point evaluation.  Must have the same shape as the dataset.

    adjustable: List[str]
        A list of columns that can be changed that you wish to analyze.

    inverse: bool, default=True
        If True, find a rule containing data points that have a different outcome from
        the current point evaluation.  Otherwise, find a rule containing data points
        that have a different outcome from the current point evaluation.

    outcome: str, default=None
        See parameter *targets*

    targets: List[str], default=None
        Optionally filters down the dataset to only the targets specified if both
        *outcome* and *targets* are defined.

    kwargs: any
        Remaining kwargs that are used in the solve (see datawhys.solver.solve)

    Returns
    -------
    rule: dict
        The conditions that the What-If Analysis found
    """

    import umap

    # TODO: find a graceful way to handle nulls/NaNs in the data
    if any(pd.isna(x) or pd.isnull(x) for x in point.values()):
        raise Exception("point cannot have any NaN or null values")

    if outcome:
        if outcome not in df.columns:
            raise Exception(f"outcome '{outcome}' does not exist")
        if is_continuous(df[outcome]):
            raise Exception(
                "outcome and targets can only be used with discrete outcomes"
            )

    if targets:
        if not outcome:
            raise Exception("outcome must be specified with targets")

        missing_targets = [x for x in targets if x not in df[outcome].unique()]

        if missing_targets:
            raise Exception(f"targets '{', '.join(missing_targets)}' do not exist")

    df_new = df.copy()
    df_new = df_new.append(point, ignore_index=True)
    df_new = df_new.dropna()
    pt_idx = df_new.shape[0] - 1

    # umap requires the df to be encoded
    encoder = DDTransformer(df_new)
    edf = utilities.encode_dataframe(df_new, encoder)

    # transform the df and add distance information
    scaler = StandardScaler()
    reducer = umap.UMAP()
    edf_scaled = scaler.fit_transform(edf)
    embedding = reducer.fit_transform(edf_scaled)

    df_new[__X_COL] = embedding[:, 0]
    df_new[__Y_COL] = embedding[:, 1]
    df_new[__DIST_COL] = (
        (df_new[__X_COL] - df_new.iloc[pt_idx][__X_COL]) ** 2
        + (df_new[__Y_COL] - df_new.iloc[pt_idx][__Y_COL]) ** 2
    ) ** 0.5
    df_new[__INV_DIST_COL] = (1.0 / df_new[__DIST_COL]).replace(
        [np.inf, -np.inf, np.nan], 0
    )

    # filter targets if given
    if outcome and targets:
        df_new = df_new[df_new[outcome].isin(targets)]

    # solve on new df to find adjustable conditions based proximity to point
    solve_outcome = __INV_DIST_COL
    solve_target = "min" if inverse else "max"
    df_new_adj = df_new[adjustable + [solve_outcome]]

    rule = solve(df_new_adj, outcome=solve_outcome, target=solve_target, **kwargs)

    return rule



from ._bias import bias
from ._correlation import correlation_groups, correlation_matrix
from ._drift import data_drift
from ._encoded import encoded_mean
from ._score import feature_signals, partial_scores, score
from ._target import target_metrics

__all__ = [
    "encoded_mean",
    "score",
    "partial_scores",
    "target_metrics",
    "correlation_matrix",
    "correlation_groups",
    "feature_signals",
    "bias",
    "data_drift",
]



from IPython.display import Image
import networkx as nx
import pygraphviz as pgv

from datawhys.utils.graph import create_rule_graph


def visualize_rule_graph(G):
    nx.nx_agraph.write_dot(G, "rules.dot")
    G = pgv.AGraph("rules.dot", directed=True)

    G.node_attr["shape"] = "box"
    G.edge_attr["color"] = "red"

    G.layout("dot")
    G.draw("graph.png")

    return Image(filename="graph.png")


def visualize_rule_tree(rule_tree):
    return visualize_rule_graph(create_rule_graph(rule_tree))



from pandas import DataFrame

from datawhys.core.series import DataWhysSeries


class DataWhysFrame(DataFrame):
    """
    A DataWhysFrame object is a pandas.DataFrame that has certain
    DataWhys specific operations and methods.

    The DataWhysFrame object is composed of 1 or more DataWhysSeries
    objects. The inner workings of the DataWhysFrame ensures that
    each variable is instantiated as a DataWhysSeries.

    For additional reference specific to the Pandas DataFrame object
    and features, please refer to the Pandas API

    **Parameters**

    data: ndarray, Iterable, dict, DataFrame, or DataWhysFrame
        Dict can contain Series, DataWhysSeries, arrays, constants,
        or list-like objects.

    **Examples**

    >>> from datawhys.core.frame import DataWhysFrame
    >>> import pandas as pd
    >>> df = pd.read_csv('path/to/file.csv')
    >>> dwf = DataWhysFrame(df)

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return DataWhysFrame

    @property
    def _constructor_sliced(self):
        return DataWhysSeries



import io
from typing import List

import numpy as np
import pandas as pd

from datawhys.api import solve_result, solve_start_file
import datawhys.datautils as datautils
from datawhys.dd_transformer import DDTransformer
from datawhys.error import (
    DataTransformationError,
    InvalidTargetError,
    NotEnoughPointsError,
)
from datawhys.metrics import score
from datawhys.utils import constants, utilities
from datawhys.utils.data import is_continuous, is_discrete


def solve(
    df: pd.DataFrame,
    outcome: str = None,
    target: str = None,
    encode=True,
    sample=True,
    random_state=None,
    return_diagnostics=False,
    **kwargs,
):
    """Run a solve without worrying about API requests.

    Parameters
    ----------
    df : pd.DataFrame
        The dataset to use for the solve

    outcome : str, default=None
        Which column to use as the outcome for the solve (sometimes referred to as a
        target feature). If None, then the first column is selected

    target: str, default=None
        The class in the outcome column to target. For a continuous outcome this should
        either be `min` or `max`. If None, then the first class or `max` is selected.

    encode: bool, default=True
        Whether or not the data should be encoded before being sent to the MB API.
        Encoding can result in additional time on client side. Disable if your data is
        largely non-sensitive.

    sample: bool, default=True
        Whether or not the data should be sampled before being sent to the MB API.
        Not pre-sampling the data can cause size limits to be reached and excessive
        solve times

    random_state : int or np.random.RandomStateInstance, default: 0
        Pseudo-random number generator to control the sampling state.
        Use an int for reproducible results across function calls.
        See the sklearn for more details.

    kwargs: any
        Remaining kwargs are sent so solve_start_file

    Returns
    -------
    rule: dict
        The conditions that the MB API found
    """
    if outcome is None:
        outcome = df.columns[0]

    if target is None:
        if df[outcome].dtype == np.object:
            target = df[outcome].iloc[0]
        else:
            target = "max"

    if target is np.nan:
        raise InvalidTargetError("Target cannot be NaN value")

    if encode:
        # Some utilities require original values
        df_orig = df.copy()
        outcome_orig = outcome
        target_orig = target

        encoder = DDTransformer(df)
        df = utilities.encode_dataframe(df, encoder)
        outcome = encoder.original_to_encoded_column(outcome)
        target = utilities.encode_value(df_orig, outcome_orig, target_orig, encoder)

    if sample:
        df = datautils.sample(
            df,
            2500,
            outcome=outcome,
            target=target,
            floor=3500,
            random_state=random_state,
        )

    _dimensionality_check(df, outcome, target)

    data = io.BytesIO()
    df.to_parquet(data)
    data.seek(0)  # Reset the pointer as `to_parquet` leaves it at the end

    task = solve_start_file(outcome=outcome, target=target, data=data, **kwargs)
    result = solve_result(id=task["id"], **kwargs)

    rule = result["rule"]
    diagnostics = result["diagnostics"]

    if encode:
        # We need to decode the rule & diagnostics if we encoded
        rule = utilities.decode_rule(rule, encoder)
        diagnostics = _decode_diagnostics(diagnostics, encoder)

    if return_diagnostics:
        return rule, diagnostics

    return rule


def _decode_diagnostics(diagnostics: dict, encoder: DDTransformer) -> dict:
    diagnostics = diagnostics.copy()

    try:
        disc_rules = diagnostics["discarded_rules"]["rules"]
        disc_rules = [utilities.decode_rule(r, encoder) for r in disc_rules]
        diagnostics["discarded_rules"]["rules"] = disc_rules
    except KeyError:
        pass

    return diagnostics


def _dimensionality_check(df: pd.DataFrame, outcome: str, target: str):
    col = df[outcome]

    if is_discrete(col):
        if len(set(col)) == 1:
            raise NotEnoughPointsError("Outcome column has only 1 modality")

        size = col[col == target].shape[0]
    else:
        if col.std() == 0:
            raise NotEnoughPointsError("Outcome column has no variance")

        size = col.shape[0]

    if size <= constants.MIN_SOLVER_SIZE:
        raise NotEnoughPointsError("Not enough points")


def _exhaustive_single(df: pd.DataFrame, outcome: str, target: str, **kwargs) -> List:
    try:
        rule = solve(df, outcome=outcome, target=target, **kwargs)
    except (NotEnoughPointsError, DataTransformationError):
        return []

    try:
        df_inverse = datautils.apply_rule(df, rule, inverse=True)
    except ValueError:
        df_inverse = df.loc[[]]

    if df_inverse.size > 0:
        return [rule] + _exhaustive_single(df_inverse, outcome, target, **kwargs)
    else:
        return []


def _exhaustive_best(df: pd.DataFrame, outcome: str, **kwargs) -> List:
    best_score, best_rule, best_target = 0, None, None

    targets = df[outcome].unique() if is_discrete(df[outcome]) else ["min", "max"]

    for target in targets:
        try:
            rule = solve(df, outcome=outcome, target=target, **kwargs)
            if rule:
                rule_score = score(df, outcome, target, rule)
            else:
                rule_score = best_score

            if best_rule is None:
                found_new_best = True
            elif is_continuous(df[outcome]) and target == "min":
                found_new_best = rule_score < best_score
            else:
                found_new_best = rule_score > best_score

            if found_new_best:
                best_rule = rule
                best_target = target
                best_score = rule_score
        except (NotEnoughPointsError, DataTransformationError):
            return []

        try:
            df_inverse = datautils.apply_rule(df, best_rule, inverse=True)
        except ValueError:
            df_inverse = df.loc[[]]

        if df_inverse.size > 0:
            return [[best_rule, best_target]] + _exhaustive_best(
                df_inverse, outcome, **kwargs
            )
        else:
            return []


def _exhaustive_any(df: pd.DataFrame, outcome: str, **kwargs) -> List:
    rule_set = []

    targets = df[outcome].unique() if is_discrete(df[outcome]) else ["min", "max"]

    for target in targets:
        try:
            rule = solve(df, outcome=outcome, target=target, **kwargs)
            rule_set.append(rule)
        except (NotEnoughPointsError, DataTransformationError):
            return []

    try:
        df_inverse = datautils.apply_rules(df, rule_set, inverse=True)
    except ValueError:
        df_inverse = df.loc[[]]

    if df_inverse.size > 0:
        return [rule_set] + _exhaustive_any(df_inverse, outcome, **kwargs)
    else:
        return []


def exhaustive_solve(
    df: pd.DataFrame,
    outcome: str = None,
    target: str = None,
    approach: str = "single",
    **solve_params,
):
    """
    Build a set of conditions that exhaustively cover the entirety of a dataset

    Parameters
    ----------
    df: Pandas DataFrame
        Data values for the exhaustive solve containing Outcome Feature
        and Independent Feature variables.

    outcome : str
        The label of the dependent variable used for learning.

    target : str (optional), default=None
        The target class of the outcome. Applicable when running a
        exhaustive solve using the 'single' approach.

    approach : str (optional), default='single'
        The approach to use for exhaustive solves. Approaches have a different
        return shape.

        - ‘single’ (default) : (n_solves,)
            Search for rules on an individual target in a dataset.
        - ‘best’ : (2, n_solves)
            Generates rules for all targets, chooses the best rule and removes
            points in that rule. Repeats for remaining data.
        - ‘any’ : (n_classes, n_solves)
            Generates rules for all targets and removes points in any rule. Repeats
            for remaining data.

    **solve_params : any
        Parameters to pass on to `solve`

    Returns
    -------
    rule_set: List
        The set of conditions that the MB API found covering all points
        in the provided dataset
    """
    if outcome is None:
        outcome = df.columns[0]

    if approach == "single":
        return _exhaustive_single(df, outcome, target, **solve_params)
    if approach == "best":
        return _exhaustive_best(df, outcome, **solve_params)
    if approach == "any":
        return _exhaustive_any(df, outcome, **solve_params)

    raise ValueError(f"approach {approach} does not exist")
